"use strict";

function _interopDefault(ex) {
  return ex && "object" == typeof ex && "default" in ex ? ex.default : ex;
}

Object.defineProperty(exports, "__esModule", {
  value: !0
});

var _objectSpread = _interopDefault(require("@babel/runtime/helpers/objectSpread")), semver = require("semver"), semver__default = _interopDefault(semver), errors = require("@changesets/errors");

function incrementVersion(release, preInfo) {
  let version = semver.inc(release.oldVersion, release.type);
  if (void 0 !== preInfo && "exit" !== preInfo.state.mode) {
    let preVersion = preInfo.preVersions.get(release.name);
    if (void 0 === preVersion) throw console.log(preInfo.preVersions.entries()), new errors.InternalError(`preVersion for ${release.name} does not exis when preState is defined`);
    version += `-${preInfo.state.tag}.${preVersion}`;
  }
  return version;
}

function getDependents(releases, workspaces, dependencyGraph, preInfo) {
  let updated = !1, pkgsToSearch = [ ...releases.values() ], pkgJsonsByName = new Map(workspaces.map(({name: name, config: config}) => [ name, config ]));
  for (;pkgsToSearch.length > 0; ) {
    const nextRelease = pkgsToSearch.shift();
    if (!nextRelease) continue;
    const pkgDependents = dependencyGraph.get(nextRelease.name);
    if (!pkgDependents) throw new Error(`Error in determining dependents - could not find package in repository: ${nextRelease.name}`);
    pkgDependents.map(dependent => {
      let type;
      const dependentPkgJSON = pkgJsonsByName.get(dependent);
      if (!dependentPkgJSON) throw new Error("Dependency map is incorrect");
      const {depTypes: depTypes, versionRange: versionRange} = getDependencyVersionRange(dependentPkgJSON, nextRelease.name);
      return depTypes.includes("peerDependencies") && "patch" !== nextRelease.type && (!releases.has(dependent) || releases.has(dependent) && "major" !== releases.get(dependent).type) ? type = "major" : releases.has(dependent) || semver__default.satisfies(incrementVersion(nextRelease, preInfo), versionRange) || (type = "patch"), 
      {
        name: dependent,
        type: type,
        pkgJSON: dependentPkgJSON
      };
    }).filter(({type: type}) => type).forEach(({name: name, type: type, pkgJSON: pkgJSON}) => {
      updated = !0;
      const existing = releases.get(name);
      if (existing && "major" === type && "major" !== existing.type) existing.type = "major", 
      pkgsToSearch.push(existing); else {
        let newDependent = {
          name: name,
          type: type,
          oldVersion: pkgJSON.version,
          changesets: []
        };
        pkgsToSearch.push(newDependent), releases.set(name, newDependent);
      }
    });
  }
  return updated;
}

function getDependencyVersionRange(dependentPkgJSON, dependencyName) {
  const DEPENDENCY_TYPES = [ "dependencies", "devDependencies", "peerDependencies", "optionalDependencies" ], dependencyVersionRange = {
    depTypes: [],
    versionRange: ""
  };
  for (const type of DEPENDENCY_TYPES) {
    const deps = dependentPkgJSON[type];
    deps && (deps[dependencyName] && (dependencyVersionRange.depTypes.push(type), dependencyVersionRange.versionRange = deps[dependencyName]));
  }
  return dependencyVersionRange;
}

function flattenReleases(changesets, workspaces) {
  let releases = new Map();
  return changesets.forEach(changeset => {
    changeset.releases.forEach(({name: name, type: type}) => {
      let release = releases.get(name), ws = workspaces.find(ws => ws.name === name);
      if (!ws) throw new Error(`Could not find package information for ${name}`);
      let {config: config} = ws;
      release ? ("minor" === release.type && "major" === type ? release.type = type : "patch" !== release.type || "major" !== type && "minor" !== type || (release.type = type), 
      release.changesets.push(changeset.id)) : release = {
        name: name,
        type: type,
        oldVersion: config.version,
        changesets: [ changeset.id ]
      }, releases.set(name, release);
    });
  }), releases;
}

function applyLinks(releases, workspaces, linked) {
  let updated = !1;
  if (!linked) return updated;
  for (let linkedPackages of linked) {
    let highestReleaseType, highestVersion, releasingLinkedPackages = [ ...releases.values() ].filter(release => linkedPackages.includes(release.name));
    if (!(releasingLinkedPackages.length < 1)) {
      for (let pkg of releasingLinkedPackages) highestReleaseType ? "major" === pkg.type ? highestReleaseType = pkg.type : "minor" === pkg.type && "major" !== highestReleaseType && (highestReleaseType = pkg.type) : highestReleaseType = pkg.type;
      for (let linkedPackage of linkedPackages) {
        let workspace = workspaces.find(workspace => workspace.name === linkedPackage);
        if (!workspace) throw console.error(`FATAL ERROR IN CHANGESETS! We were unable to version for linked package: ${linkedPackage} in linkedPackages: ${linkedPackages.toString()}`), 
        new Error("fatal: could not resolve linked packages");
        (void 0 === highestVersion || semver__default.gt(workspace.config.version, highestVersion)) && (highestVersion = workspace.config.version);
      }
      if (!highestVersion || !highestReleaseType) throw new Error("Large internal changesets error in calculating linked versions. Please contact the maintainers");
      for (let linkedPackage of releasingLinkedPackages) linkedPackage.type !== highestReleaseType && (updated = !0, 
      linkedPackage.type = highestReleaseType), linkedPackage.oldVersion !== highestVersion && (updated = !0, 
      linkedPackage.oldVersion = highestVersion);
    }
  }
  return updated;
}

function getPreVersion(version) {
  let parsed = semver.parse(version), preVersion = void 0 === parsed.prerelease[1] ? -1 : parsed.prerelease[1];
  if ("number" != typeof preVersion) throw new errors.InternalError("preVersion is not a number");
  return ++preVersion;
}

function assembleReleasePlan(changesets, workspaces, dependentsGraph, config, preState) {
  let updatedPreState = void 0 === preState ? void 0 : _objectSpread({}, preState, {
    initialVersions: _objectSpread({}, preState.initialVersions)
  }), workspacesByName = new Map(workspaces.map(x => [ x.name, x ]));
  if (void 0 !== updatedPreState) {
    for (let workspace of workspaces) void 0 === updatedPreState.initialVersions[workspace.name] && (updatedPreState.initialVersions[workspace.name] = workspace.config.version);
    if ("exit" !== updatedPreState.mode) {
      let usedChangesetIds = new Set(updatedPreState.changesets);
      updatedPreState.changesets = changesets.map(x => x.id), changesets = changesets.filter(changeset => !usedChangesetIds.has(changeset.id));
    }
  }
  let releases = flattenReleases(changesets, workspaces), preVersions = new Map();
  if (void 0 !== updatedPreState) {
    for (let [name, release] of releases) releases.set(name, _objectSpread({}, release, {
      oldVersion: updatedPreState.initialVersions[name]
    }));
    if ("exit" !== updatedPreState.mode) {
      for (let workspace of workspaces) preVersions.set(workspace.name, getPreVersion(workspace.config.version));
      for (let linkedGroup of config.linked) {
        let highestPreVersion = 0;
        for (let linkedPackage of linkedGroup) highestPreVersion = Math.max(getPreVersion(workspacesByName.get(linkedPackage).config.version), highestPreVersion);
        for (let linkedPackage of linkedGroup) preVersions.set(linkedPackage, highestPreVersion);
      }
    }
  }
  if (void 0 !== updatedPreState && "exit" === updatedPreState.mode) for (let workspace of workspaces) semver.parse(workspace.config.version).prerelease.length && (releases.has(workspace.name) || releases.set(workspace.name, {
    type: "patch",
    name: workspace.name,
    changesets: [],
    oldVersion: updatedPreState.initialVersions[workspace.name]
  }));
  let preInfo = void 0 === updatedPreState ? void 0 : {
    state: updatedPreState,
    preVersions: preVersions
  }, releaseObjectValidated = !1;
  for (;!1 === releaseObjectValidated; ) {
    let dependentAdded = getDependents(releases, workspaces, dependentsGraph, preInfo);
    releaseObjectValidated = !applyLinks(releases, workspaces, config.linked) && !dependentAdded;
  }
  return {
    changesets: changesets,
    releases: [ ...releases.values() ].map(incompleteRelease => _objectSpread({}, incompleteRelease, {
      newVersion: incrementVersion(incompleteRelease, preInfo)
    })),
    preState: updatedPreState
  };
}

exports.default = assembleReleasePlan;
